@@ -52,14 +52,25 @@
 #define DEFAULT_IPC_COMM_TIMEOUT 40
 #define DEFAULT_OUTPUT_BUFFERSIZE 65536
 #define DEFAULT_MAX_REQUESTS_PER_PROCESS 0
-#define DEFAULT_MAX_REQUEST_LEN (1024*1024*1024)    /* 1G */
-#define DEFAULT_MAX_MEM_REQUEST_LEN (1024*64)   /* 64k */
+/* by default, allow spooling of request bodies up to
+ * 128k (first 64k in memory)
+ */
+#define DEFAULT_MAX_REQUEST_LEN (1024*128)
+#define DEFAULT_MAX_MEM_REQUEST_LEN (1024*64)
 #define DEFAULT_WRAPPER_KEY "ALL"
 #define WRAPPER_FLAG_VIRTUAL "virtual"
 
 void *create_fcgid_server_config(apr_pool_t * p, server_rec * s)
 {
     fcgid_server_conf *config = apr_pcalloc(p, sizeof(*config));
+    static int vhost_id = 0;
+
+    /* allow vhost comparison even when some mass-vhost module
+     * makes a copy of the server_rec to override docroot or
+     * other such settings
+     */
+    ++vhost_id;
+    config->vhost_id = vhost_id;
 
     if (!s->is_virtual) {
         config->busy_scan_interval = DEFAULT_BUSY_SCAN_INTERVAL;
@@ -621,11 +632,12 @@ const char *set_authenticator_info(cmd_parms * cmd, void *config,
     dirconfig->authenticator_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->authenticator_info));
-    apr_cpystrn(dirconfig->authenticator_info->path, authenticator,
-                _POSIX_PATH_MAX);
+    dirconfig->authenticator_info->cgipath =
+        apr_pstrdup(cmd->pool, authenticator);
+    dirconfig->authenticator_info->cmdline =
+        dirconfig->authenticator_info->cgipath;
     dirconfig->authenticator_info->inode = finfo.inode;
     dirconfig->authenticator_info->deviceid = finfo.device;
-    dirconfig->authenticator_info->share_group_id = (apr_size_t) - 1;
     return NULL;
 }
 
@@ -639,7 +651,7 @@ const char *set_authenticator_authoritative(cmd_parms * cmd,
     return NULL;
 }
 
-auth_conf *get_authenticator_info(request_rec * r, int *authoritative)
+fcgid_cmd_conf *get_authenticator_info(request_rec * r, int *authoritative)
 {
     fcgid_dir_conf *config =
         ap_get_module_config(r->per_dir_config, &fcgid_module);
@@ -669,11 +681,12 @@ const char *set_authorizer_info(cmd_parms * cmd, void *config,
     dirconfig->authorizer_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->authorizer_info));
-    apr_cpystrn(dirconfig->authorizer_info->path, authorizer,
-                _POSIX_PATH_MAX);
+    dirconfig->authorizer_info->cgipath =
+        apr_pstrdup(cmd->pool, authorizer);
+    dirconfig->authorizer_info->cmdline =
+        dirconfig->authorizer_info->cgipath;
     dirconfig->authorizer_info->inode = finfo.inode;
     dirconfig->authorizer_info->deviceid = finfo.device;
-    dirconfig->authorizer_info->share_group_id = (apr_size_t) - 1;
     return NULL;
 }
 
@@ -687,7 +700,7 @@ const char *set_authorizer_authoritative(cmd_parms * cmd,
     return NULL;
 }
 
-auth_conf *get_authorizer_info(request_rec * r, int *authoritative)
+fcgid_cmd_conf *get_authorizer_info(request_rec * r, int *authoritative)
 {
     fcgid_dir_conf *config =
         ap_get_module_config(r->per_dir_config, &fcgid_module);
@@ -717,10 +730,12 @@ const char *set_access_info(cmd_parms * cmd, void *config,
     dirconfig->access_info =
         apr_pcalloc(cmd->server->process->pconf,
                     sizeof(*dirconfig->access_info));
-    apr_cpystrn(dirconfig->access_info->path, access, _POSIX_PATH_MAX);
+    dirconfig->access_info->cgipath =
+        apr_pstrdup(cmd->pool, access);
+    dirconfig->access_info->cmdline = 
+        dirconfig->access_info->cgipath;
     dirconfig->access_info->inode = finfo.inode;
     dirconfig->access_info->deviceid = finfo.device;
-    dirconfig->access_info->share_group_id = (apr_size_t) - 1;
     return NULL;
 }
 
@@ -734,7 +749,7 @@ const char *set_access_authoritative(cmd_parms * cmd,
     return NULL;
 }
 
-auth_conf *get_access_info(request_rec * r, int *authoritative)
+fcgid_cmd_conf *get_access_info(request_rec * r, int *authoritative)
 {
     fcgid_dir_conf *config =
         ap_get_module_config(r->per_dir_config, &fcgid_module);
@@ -747,14 +762,6 @@ auth_conf *get_access_info(request_rec * r, int *authoritative)
     return NULL;
 }
 
-typedef struct {
-    apr_hash_t *wrapper_id_hash;
-    apr_size_t cur_id;
-} wrapper_id_info;
-
-/* FIXME thread safety issues when FcgidWrapper is used in .htaccess;
- * see use of pconf
- */
 const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
                                const char *wrapper_cmdline,
                                const char *extension,
@@ -763,11 +770,7 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
     const char *path, *tmp;
     apr_status_t rv;
     apr_finfo_t finfo;
-    const char *userdata_key = "fcgid_wrapper_id";
-    wrapper_id_info *id_info;
-    apr_size_t *wrapper_id;
-    fcgid_wrapper_conf *wrapper = NULL;
-    apr_pool_t *wrapper_conf_pool = cmd->server->process->pconf; /* bad */
+    fcgid_cmd_conf *wrapper = NULL;
     fcgid_dir_conf *config = (fcgid_dir_conf *) dirconfig;
 
     /* Sanity checks */
@@ -786,36 +789,6 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
             || ap_strchr_c(extension, '/') || ap_strchr_c(extension, '\\')))
         return "Invalid wrapper file extension";
 
-    /* Get wrapper_id hash from user data */
-    {
-        void *id_info_vp;
-        apr_pool_userdata_get(&id_info_vp, userdata_key,
-                              cmd->server->process->pool);
-        id_info = id_info_vp;
-    }
-
-    if (!id_info) {
-        id_info =
-            apr_pcalloc(cmd->server->process->pool, sizeof(*id_info));
-        id_info->wrapper_id_hash =
-            apr_hash_make(cmd->server->process->pool);
-        apr_pool_userdata_set((const void *) id_info, userdata_key,
-                              apr_pool_cleanup_null,
-                              cmd->server->process->pool);
-    }
-    /* Get wrapper_id for wrapper_cmdline */
-    if ((wrapper_id =
-         apr_hash_get(id_info->wrapper_id_hash, wrapper_cmdline,
-                      strlen(wrapper_cmdline))) == NULL) {
-        wrapper_id =
-            apr_pcalloc(cmd->server->process->pool, sizeof(*wrapper_id));
-        *wrapper_id = id_info->cur_id++;
-        apr_hash_set(id_info->wrapper_id_hash, wrapper_cmdline,
-                     strlen(wrapper_cmdline), wrapper_id);
-    }
-
-    wrapper = apr_pcalloc(wrapper_conf_pool, sizeof(*wrapper));
-
     /* Get wrapper path */
     tmp = wrapper_cmdline;
     path = ap_getword_white(cmd->temp_pool, &tmp);
@@ -828,14 +801,21 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
         return missing_file_msg(cmd->pool, "Wrapper", path, rv);
     }
 
-    wrapper->exe = apr_pstrdup(wrapper_conf_pool, path);
-    /* FIXME no need to embed in structure (subject to correct pool usage) */
-    apr_cpystrn(wrapper->args, wrapper_cmdline, _POSIX_PATH_MAX);
+    wrapper = apr_pcalloc(cmd->pool, sizeof(*wrapper));
+
+    if (strlen(path) >= FCGID_PATH_MAX) {
+        return "Executable path length exceeds compiled-in limit";
+    }
+    wrapper->cgipath = apr_pstrdup(cmd->pool, path);
+
+    if (strlen(wrapper_cmdline) >= FCGID_CMDLINE_MAX) {
+        return "Command line length exceeds compiled-in limit";
+    }
+    wrapper->cmdline = apr_pstrdup(cmd->pool, wrapper_cmdline);
+
     wrapper->inode = finfo.inode;
     wrapper->deviceid = finfo.device;
-    wrapper->share_group_id = *wrapper_id;
     wrapper->virtual = (virtual != NULL && !strcasecmp(virtual, WRAPPER_FLAG_VIRTUAL));
-    (*wrapper_id)++;
 
     if (extension == NULL)
         extension = DEFAULT_WRAPPER_KEY;
@@ -848,10 +828,10 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
     return NULL;
 }
 
-fcgid_wrapper_conf *get_wrapper_info(const char *cgipath, request_rec * r)
+fcgid_cmd_conf *get_wrapper_info(const char *cgipath, request_rec * r)
 {
     const char *extension;
-    fcgid_wrapper_conf *wrapper;
+    fcgid_cmd_conf *wrapper;
     fcgid_dir_conf *config =
         ap_get_module_config(r->per_dir_config, &fcgid_module);
 
