@@ -28,6 +28,7 @@
 #include "fcgid_conf.h"
 #include "fcgid_proctbl.h"
 #include "fcgid_spawn_ctl.h"
+#include "fcgid_mutex.h"
 #include <unistd.h>
 
 #if MODULE_MAGIC_NUMBER_MAJOR >= 20090209
@@ -37,9 +38,17 @@
 #if MODULE_MAGIC_NUMBER_MAJOR < 20081201
 #define ap_unixd_config unixd_config
 #define ap_unixd_setup_child unixd_setup_child
-#define ap_unixd_set_global_mutex_perms unixd_set_global_mutex_perms
 #endif
 
+/* The APR other-child API doesn't tell us how the daemon exited
+ * (SIGSEGV vs. exit(1)).  The other-child maintenance function
+ * needs to decide whether to restart the daemon after a failure
+ * based on whether or not it exited due to a fatal startup error
+ * or something that happened at steady-state.  This exit status
+ * is unlikely to collide with exit signals.
+ */
+#define DAEMON_STARTUP_ERROR 254
+
 static apr_status_t create_process_manager(server_rec * main_server,
                                            apr_pool_t * configpool);
 
@@ -50,7 +59,8 @@ static apr_file_t *g_pm_write_pipe = NULL;
 static apr_file_t *g_ap_write_pipe = NULL;
 static apr_file_t *g_ap_read_pipe = NULL;
 static apr_global_mutex_t *g_pipelock = NULL;
-char g_pipelock_name[L_tmpnam];
+static const char *g_pipelock_name;
+static const char *g_pipelock_mutex_type = "fcgid-pipe";
 
 static int volatile g_caughtSigTerm = 0;
 static pid_t g_pm_pid;
@@ -120,18 +130,33 @@ static void fcgid_maint(int reason, void *data, apr_wait_t status)
         apr_proc_other_child_unregister(data);
         if (ap_mpm_query(AP_MPMQ_MPM_STATE, &mpm_state) == APR_SUCCESS
             && mpm_state != AP_MPMQ_STOPPING) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
-                         "mod_fcgid: fcgid process manager died, restarting the server");
+            if (status == DAEMON_STARTUP_ERROR) {
+                ap_log_error(APLOG_MARK, APLOG_CRIT, 0, NULL,
+                             "mod_fcgid: fcgid process manager failed to initialize; "
+                             "stopping httpd");
+                /* mod_fcgid requests will hang due to lack of a process manager;
+                 * try to terminate httpd
+                 */
+                kill(getpid(), SIGTERM);
+                
+            }
+            else {
+                ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
+                             "mod_fcgid: fcgid process manager died, restarting the server");
 
-            /* HACK: I can't just call create_process_manager() to
-               restart a process manager, because it will use the dirty
-               share memory, I have to kill myself a SIGHUP, to make
-               a clean restart */
-            if (kill(getpid(), SIGHUP) < 0) {
-                ap_log_error(APLOG_MARK, APLOG_EMERG,
-                             apr_get_os_error(), NULL,
-                             "mod_fcgid: can't send SIGHUP to self");
-                exit(0);
+                /* HACK: I can't just call create_process_manager() to
+                   restart a process manager, because it will use the dirty
+                   share memory, I have to kill myself a SIGHUP, to make
+                   a clean restart */
+                /* FIXME: This is the httpd parent; mod_fcgid is doing a hard
+                 * restart of the server!
+                 */
+                if (kill(getpid(), SIGHUP) < 0) {
+                    ap_log_error(APLOG_MARK, APLOG_EMERG,
+                                 apr_get_os_error(), NULL,
+                                 "mod_fcgid: can't send SIGHUP to self");
+                    exit(0);
+                }
             }
         }
         break;
@@ -141,6 +166,9 @@ static void fcgid_maint(int reason, void *data, apr_wait_t status)
     case APR_OC_REASON_LOST:
         apr_proc_other_child_unregister(data);
         /* It hack here too, a note above */
+        /* FIXME: This is the httpd parent; mod_fcgid is doing a hard
+         * restart of the server!
+         */
         if (kill(getpid(), SIGHUP) < 0) {
             ap_log_error(APLOG_MARK, APLOG_EMERG,
                          apr_get_os_error(), NULL,
@@ -218,7 +246,7 @@ static int suexec_setup_child(void)
     /* Only try to switch if we're running as root */
     if (!geteuid() && (seteuid(ap_unixd_config.user_id) == -1)) {
         ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-                     "setuid: unable to change to uid: %ld",
+                     "seteuid: unable to change to uid %ld",
                      (long) ap_unixd_config.user_id);
         exit(-1);
     }
@@ -240,17 +268,25 @@ create_process_manager(server_rec * main_server, apr_pool_t * configpool)
                      "mod_fcgid: Process manager %" APR_PID_T_FMT  " started", getpid());
 
         if ((rv = init_signal(main_server)) != APR_SUCCESS) {
-            ap_log_error(APLOG_MARK, APLOG_EMERG, 0, main_server,
+            ap_log_error(APLOG_MARK, APLOG_EMERG, rv, main_server,
                          "mod_fcgid: can't install signal handler, exiting now");
-            exit(1);
+            exit(DAEMON_STARTUP_ERROR);
         }
 
-        /* if running as root, switch to configured user */
+        /* If running as root, switch to configured user.
+         *
+         * When running children via suexec, only the effective uid is
+         * switched, so that the PM can return to euid 0 to kill child
+         * processes.
+         *
+         * When running children as the configured user, the real uid
+         * is switched.
+         */
         if (ap_unixd_config.suexec_enabled) {
             if (getuid() != 0) {
                 ap_log_error(APLOG_MARK, APLOG_EMERG, 0, main_server,
                              "mod_fcgid: current user is not root while suexec is enabled, exiting now");
-                exit(1);
+                exit(DAEMON_STARTUP_ERROR);
             }
             suexec_setup_child();
         } else
@@ -301,6 +337,12 @@ procmgr_child_init(server_rec * main_server, apr_pool_t * configpool)
     return APR_SUCCESS;
 }
 
+apr_status_t procmgr_pre_config(apr_pool_t *p, apr_pool_t *plog,
+                                apr_pool_t *ptemp)
+{
+    return fcgid_mutex_register(g_pipelock_mutex_type, p);
+}
+
 apr_status_t
 procmgr_post_config(server_rec * main_server, apr_pool_t * configpool)
 {
@@ -364,18 +406,10 @@ procmgr_post_config(server_rec * main_server, apr_pool_t * configpool)
     }
 
     /* Create mutex for pipe reading and writing */
-    if ((rv =
-         apr_global_mutex_create(&g_pipelock, tmpnam(g_pipelock_name),
-                                 APR_LOCK_DEFAULT,
-                                 main_server->process->pconf)) !=
-        APR_SUCCESS) {
-        ap_log_error(APLOG_MARK, APLOG_EMERG, rv, main_server,
-                     "mod_fcgid: Can't create global pipe mutex");
-        exit(1);
-    }
-    if ((rv = ap_unixd_set_global_mutex_perms(g_pipelock)) != APR_SUCCESS) {
-        ap_log_error(APLOG_MARK, APLOG_EMERG, rv, main_server,
-                     "mod_fcgid: Can't set global pipe mutex perms");
+    rv = fcgid_mutex_create(&g_pipelock, &g_pipelock_name,
+                            g_pipelock_mutex_type,
+                            main_server->process->pconf, main_server);
+    if (rv != APR_SUCCESS) {
         exit(1);
     }
 
@@ -384,12 +418,11 @@ procmgr_post_config(server_rec * main_server, apr_pool_t * configpool)
 }
 
 void procmgr_init_spawn_cmd(fcgid_command * command, request_rec * r,
-                            const char *argv0, dev_t deviceid,
-                            apr_ino_t inode, apr_size_t share_grp_id)
+                            fcgid_cmd_conf *cmd_conf)
 {
     ap_unix_identity_t *ugid;
-    fcgid_wrapper_conf *wrapperconf;
-    const char *cmd_to_spawn;
+    fcgid_server_conf *sconf =
+        ap_get_module_config(r->server->module_config, &fcgid_module);
 
     memset(command, 0, sizeof(*command));
 
@@ -404,26 +437,21 @@ void procmgr_init_spawn_cmd(fcgid_command * command, request_rec * r,
         command->userdir = 0;
     }
 
-    apr_cpystrn(command->cgipath, argv0, _POSIX_PATH_MAX);
-    command->deviceid = deviceid;
-    command->inode = inode;
-    command->share_grp_id = share_grp_id;
-    command->virtualhost = r->server->server_hostname;
+    /* no truncation should ever occur */
+    AP_DEBUG_ASSERT(sizeof command->cgipath > strlen(cmd_conf->cgipath));
+    apr_cpystrn(command->cgipath, cmd_conf->cgipath, sizeof command->cgipath);
+    AP_DEBUG_ASSERT(sizeof command->cmdline > strlen(cmd_conf->cmdline));
+    apr_cpystrn(command->cmdline, cmd_conf->cmdline, sizeof command->cmdline);
 
-    /* Update fcgid_command with wrapper info */
-    command->wrapperpath[0] = '\0';
-    if ((wrapperconf = get_wrapper_info(argv0, r))) {
-        apr_cpystrn(command->wrapperpath, wrapperconf->args, _POSIX_PATH_MAX);
-        command->deviceid = wrapperconf->deviceid;
-        command->inode = wrapperconf->inode;
-        command->share_grp_id = wrapperconf->share_group_id;
-        cmd_to_spawn = command->wrapperpath;
-    }
-    else {
-        cmd_to_spawn = command->cgipath;
+    command->deviceid = cmd_conf->deviceid;
+    command->inode = cmd_conf->inode;
+    command->vhost_id = sconf->vhost_id;
+    if (r->server->server_hostname) {
+        apr_cpystrn(command->server_hostname, r->server->server_hostname,
+                    sizeof command->server_hostname);
     }
 
-    get_cmd_options(r, cmd_to_spawn, &command->cmdopts);
+    get_cmd_options(r, command->cgipath, &command->cmdopts, &command->cmdenv);
 }
 
 apr_status_t procmgr_post_spawn_cmd(fcgid_command * command,
@@ -433,13 +461,9 @@ apr_status_t procmgr_post_spawn_cmd(fcgid_command * command,
     char notifybyte;
     apr_size_t nbytes = sizeof(*command);
 
-    /* Sanity check first */
-    if (g_caughtSigTerm || !g_ap_write_pipe)
-        return APR_SUCCESS;
-
     /* Get the global mutex before posting the request */
     if ((rv = apr_global_mutex_lock(g_pipelock)) != APR_SUCCESS) {
-        ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
+        ap_log_rerror(APLOG_MARK, APLOG_EMERG, rv, r,
                       "mod_fcgid: can't get pipe mutex");
         exit(0);
     }
@@ -463,7 +487,7 @@ apr_status_t procmgr_post_spawn_cmd(fcgid_command * command,
 
     /* Release the lock */
     if ((rv = apr_global_mutex_unlock(g_pipelock)) != APR_SUCCESS) {
-        ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
+        ap_log_rerror(APLOG_MARK, APLOG_EMERG, rv, r,
                       "mod_fcgid: can't release pipe mutex");
         exit(0);
     }
