@@ -128,37 +128,29 @@ procmgr_post_config(server_rec * main_server, apr_pool_t * pconf)
 }
 
 void procmgr_init_spawn_cmd(fcgid_command * command, request_rec * r,
-                            const char *argv0, dev_t deviceid,
-                            apr_ino_t inode, apr_size_t share_grp_id)
+                            fcgid_cmd_conf *cmd_conf)
 {
-    fcgid_wrapper_conf *wrapperconf;
-    const char *cmd_to_spawn;
+    fcgid_server_conf *sconf =
+        ap_get_module_config(r->server->module_config, &fcgid_module);
 
     memset(command, 0, sizeof(*command));
 
-    apr_cpystrn(command->cgipath, argv0, _POSIX_PATH_MAX);
-    command->deviceid = deviceid;
-    command->inode = inode;
-    command->share_grp_id = share_grp_id;
+    /* no truncation should ever occur */
+    AP_DEBUG_ASSERT(sizeof command->cgipath > strlen(cmd_conf->cgipath));
+    apr_cpystrn(command->cgipath, cmd_conf->cgipath, sizeof command->cgipath);
+    AP_DEBUG_ASSERT(sizeof command->cmdline > strlen(cmd_conf->cmdline));
+    apr_cpystrn(command->cmdline, cmd_conf->cmdline, sizeof command->cmdline);
+
     command->uid = (uid_t) - 1;
     command->gid = (gid_t) - 1;
     command->userdir = 0;
-    command->virtualhost = r->server->server_hostname;
-
-    /* Update fcgid_command with wrapper info */
-    command->wrapper_cmdline[0] = '\0';
-    if ((wrapperconf = get_wrapper_info(argv0, r))) {
-        apr_cpystrn(command->wrapper_cmdline, wrapperconf->args, _POSIX_PATH_MAX);
-        command->deviceid = wrapperconf->deviceid;
-        command->inode = wrapperconf->inode;
-        command->share_grp_id = wrapperconf->share_group_id;
-        cmd_to_spawn = wrapperconf->exe;
-    }
-    else {
-        cmd_to_spawn = command->cgipath;
+    command->vhost_id = sconf->vhost_id;
+    if (r->server->server_hostname) {
+        apr_cpystrn(command->server_hostname, r->server->server_hostname,
+                    sizeof command->server_hostname);
     }
 
-    get_cmd_options(r, cmd_to_spawn, &command->cmdopts, &command->cmdenv);
+    get_cmd_options(r, command->cgipath, &command->cmdopts, &command->cmdenv);
 }
 
 apr_status_t procmgr_post_spawn_cmd(fcgid_command * command,
