@@ -91,7 +91,8 @@ static apr_status_t ap_unix_create_privileged_process(apr_proc_t *newproc,
 
     if (ugid->userdir) {
         execuser = apr_psprintf(p, "~%ld", (long) ugid->uid);
-    } else {
+    }
+    else {
         execuser = apr_psprintf(p, "%ld", (long) ugid->uid);
     }
     execgroup = apr_psprintf(p, "%ld", (long) ugid->gid);
@@ -160,16 +161,38 @@ static apr_status_t socket_file_cleanup(void *theprocnode)
     return APR_SUCCESS;
 }
 
+static void log_setid_failure(const char *proc_type,
+                              const char *id_type,
+                              uid_t user_id)
+{
+    char errno_desc[120];
+    char errmsg[240];
+
+    apr_strerror(errno, errno_desc, sizeof errno_desc);
+    apr_snprintf(errmsg, sizeof errmsg,
+                 "(%d)%s: %s unable to set %s to %ld\n",
+                 errno, errno_desc, proc_type, id_type, (long)user_id);
+    write(STDERR_FILENO, errmsg, strlen(errmsg));
+}
+
+/* When suexec is enabled, this runs in the forked child
+ * process prior to exec().
+ */
 static apr_status_t exec_setuid_cleanup(void *dummy)
 {
-    seteuid(0);
-    setuid(ap_unixd_config.user_id);
+    if (seteuid(0) == -1) {
+        log_setid_failure("mod_fcgid child", "effective uid", 0);
+        _exit(1);
+    }
+    if (setuid(ap_unixd_config.user_id) == -1) {
+        log_setid_failure("mod_fcgid child", "uid", ap_unixd_config.user_id);
+        _exit(1);
+    }
     return APR_SUCCESS;
 }
 
-apr_status_t
-proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
-                   fcgid_procnode * procnode)
+apr_status_t proc_spawn_process(const char *cmdline, fcgid_proc_info *procinfo,
+                                fcgid_procnode *procnode)
 {
     server_rec *main_server = procinfo->main_server;
     fcgid_server_conf *sconf = ap_get_module_config(main_server->module_config,
@@ -181,15 +204,14 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
     char **proc_environ;
     struct sockaddr_un unix_addr;
     apr_procattr_t *procattr = NULL;
-    int argc;
-    const char *wargv[APACHE_ARG_MAX];
+    int argc, len;
+    const char *wargv[APACHE_ARG_MAX + 1];
     const char *word; /* For wrapper */
     const char *tmp;
-    const char *argv[2];
 
     /* Build wrapper args */
     argc = 0;
-    tmp = lpszwapper;
+    tmp = cmdline;
     while (1) {
         word = ap_getword_white(procnode->proc_pool, &tmp);
         if (word == NULL || *word == '\0')
@@ -205,17 +227,31 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
      */
 
     /* Generate a UNIX domain socket file path */
-    /* XXX It's nothing I can do if the socket dir is too long... */
     memset(&unix_addr, 0, sizeof(unix_addr));
     unix_addr.sun_family = AF_UNIX;
-    apr_snprintf(unix_addr.sun_path, sizeof(unix_addr.sun_path) - 1,
-                 "%s/%" APR_PID_T_FMT ".%d", sconf->sockname_prefix,
-                 getpid(), g_process_counter++);
-    strncpy(procnode->socket_path, unix_addr.sun_path,
-            sizeof(procnode->socket_path) - 1);
-    strncpy(procnode->executable_path,
-            (lpszwapper != NULL && lpszwapper[0] != '\0') ? wargv[0] : procinfo->cgipath,
-            sizeof(procnode->executable_path) - 1);
+    len = apr_snprintf(unix_addr.sun_path, sizeof(unix_addr.sun_path),
+                       "%s/%" APR_PID_T_FMT ".%d", sconf->sockname_prefix,
+                       getpid(), g_process_counter++);
+
+    /* check for truncation of the socket path
+     *
+     * cheap but overly zealous check for sun_path overflow: if length of
+     * prepared string is at the limit, assume truncation
+     */
+    if (len + 1 == sizeof(unix_addr.sun_path)
+        || len >= sizeof procnode->socket_path) {
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, main_server,
+                     "mod_fcgid: socket path length exceeds compiled-in limits");
+        return APR_EGENERAL;
+    }
+
+    apr_cpystrn(procnode->socket_path, unix_addr.sun_path,
+                sizeof(procnode->socket_path));
+
+    /* truncation already checked for in handler or FcgidWrapper parser */
+    AP_DEBUG_ASSERT(strlen(wargv[0]) < sizeof(procnode->executable_path));
+    apr_cpystrn(procnode->executable_path, wargv[0],
+                sizeof(procnode->executable_path));
 
     /* Unlink the file just in case */
     unlink(unix_addr.sun_path);
@@ -227,12 +263,16 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
         return errno;
     }
 
-    /* Unlink it when process exit */
+    /* Register cleanups to
+     * 1. Unlink the socket when the process exits
+     * 2. (suexec mode only, in the child cleanup) Switch to the configured uid
+     */
     if (ap_unixd_config.suexec_enabled) {
         apr_pool_cleanup_register(procnode->proc_pool,
                                   procnode, socket_file_cleanup,
                                   exec_setuid_cleanup);
-    } else {
+    }
+    else {
         apr_pool_cleanup_register(procnode->proc_pool,
                                   procnode, socket_file_cleanup,
                                   apr_pool_cleanup_null);
@@ -254,7 +294,8 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
     /* IPC directory permissions are safe, but avoid confusion */
     /* Not all flavors of unix use the current umask for AF_UNIX perms */
 
-    rv = apr_file_perms_set(unix_addr.sun_path, APR_FPROT_UREAD|APR_FPROT_UWRITE|APR_FPROT_UEXECUTE);
+    rv = apr_file_perms_set(unix_addr.sun_path,
+                            APR_FPROT_UREAD|APR_FPROT_UWRITE|APR_FPROT_UEXECUTE);
     if (rv != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_CRIT, rv, main_server,
                      "mod_fcgid: Couldn't set permissions on unix domain socket %s",
@@ -314,30 +355,20 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
     }
 
     /* Prepare the fork */
-    if (
-   (rv =
-    apr_procattr_create(&procattr, procnode->proc_pool)) != APR_SUCCESS
-|| (rv =
-    apr_procattr_child_err_set(procattr,
-                               procinfo->main_server->error_log,
-                               NULL)) != APR_SUCCESS
-|| (rv =
-    apr_procattr_child_out_set(procattr,
-                               procinfo->main_server->error_log,
-                               NULL)) != APR_SUCCESS
-|| (rv =
-    apr_procattr_dir_set(procattr,
-                         ap_make_dirstr_parent(procnode->proc_pool,
-                                               (lpszwapper != NULL
-                                                && lpszwapper[0] !=
-                                                '\0') ? wargv[0] :
-                                               procinfo->cgipath))) !=
-APR_SUCCESS
-|| (rv = apr_procattr_cmdtype_set(procattr, APR_PROGRAM)) != APR_SUCCESS
-|| (rv =
-    apr_os_file_put(&file, &unix_socket, 0,
-                    procnode->proc_pool)) != APR_SUCCESS
-|| (rv = apr_procattr_child_in_set(procattr, file, NULL)) != APR_SUCCESS) {
+    if ((rv = apr_procattr_create(&procattr, procnode->proc_pool)) != APR_SUCCESS
+        || (rv = apr_procattr_child_err_set(procattr,
+                                            procinfo->main_server->error_log,
+                                            NULL)) != APR_SUCCESS
+        || (rv = apr_procattr_child_out_set(procattr,
+                                            procinfo->main_server->error_log,
+                                            NULL)) != APR_SUCCESS
+        || (rv = apr_procattr_dir_set(procattr,
+                                      ap_make_dirstr_parent(procnode->proc_pool,
+                                                            wargv[0]))) != APR_SUCCESS
+        || (rv = apr_procattr_cmdtype_set(procattr, APR_PROGRAM)) != APR_SUCCESS
+        || (rv = apr_os_file_put(&file, &unix_socket, 0,
+                                 procnode->proc_pool)) != APR_SUCCESS
+        || (rv = apr_procattr_child_in_set(procattr, file, NULL)) != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
                      "mod_fcgid: couldn't set child process attributes: %s",
                      unix_addr.sun_path);
@@ -350,96 +381,70 @@ APR_SUCCESS
      * for it's a share memory address, both parent and child process may modify 
      * procnode->proc_id->pid, so sometimes it's 0 and sometimes it's >0
      */
-    if (lpszwapper != NULL && lpszwapper[0] != '\0') {
-        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, procinfo->main_server,
-                     "mod_fcgid: call %s with wrapper %s",
-                     procinfo->cgipath, lpszwapper);
-        if ((rv =
-             fcgid_create_privileged_process(&tmpproc,
-                                             wargv[0], wargv,
-                                             (const char *const *)
-                                             proc_environ, procattr,
-                                             procinfo,
-                                             procnode->proc_pool)) !=
-            APR_SUCCESS) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
-                         "mod_fcgid: can't create wrapper process for %s",
-                         procinfo->cgipath);
-            close(unix_socket);
-            procnode->proc_id = tmpproc;
-            return rv;
-        }
-    } else {
-        argv[0] = procinfo->cgipath;
-        argv[1] = NULL;
-        if ((rv =
-             fcgid_create_privileged_process(&tmpproc,
-                                             procinfo->cgipath,
-                                             argv,
-                                             (const char *const *)
-                                             proc_environ, procattr,
-                                             procinfo,
-                                             procnode->proc_pool)) !=
-            APR_SUCCESS) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
-                         "mod_fcgid: can't create process");
-            close(unix_socket);
-            procnode->proc_id = tmpproc;
-            return rv;
-        }
+    rv = fcgid_create_privileged_process(&tmpproc, wargv[0], wargv,
+                                         (const char *const *)proc_environ,
+                                         procattr, procinfo,
+                                         procnode->proc_pool);
+
+    if (ap_unixd_config.suexec_enabled) {
+        /* Prior to creating the child process, a child cleanup was registered
+         * to switch the uid in the child.  No-op the child cleanup for this
+         * pool so that it won't run again as other child processes are created.
+         * (The cleanup will be registered for the pool associated with those
+         * processes too.)
+         */
+        apr_pool_child_cleanup_set(procnode->proc_pool, procnode,
+                                   socket_file_cleanup, apr_pool_cleanup_null);
     }
 
     /* Close socket before try to connect to it */
     close(unix_socket);
     procnode->proc_id = tmpproc;
 
-    return APR_SUCCESS;
+    if (rv != APR_SUCCESS) {
+        ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
+                     "mod_fcgid: can't run %s", wargv[0]);
+    }
+
+    return rv;
 }
 
-apr_status_t
-proc_kill_gracefully(fcgid_procnode * procnode, server_rec * main_server)
+static apr_status_t proc_kill_internal(fcgid_procnode *procnode, int sig)
 {
-
-    /* su as root before sending kill signal, for suEXEC */
+    /* su as root before sending signal, for suEXEC */
     apr_status_t rv;
 
     if (ap_unixd_config.suexec_enabled && seteuid(0) != 0) {
 
-        /* It's fatal error */
-        kill(getpid(), SIGTERM);
-        return APR_EACCES;
+        /* can't gain privileges to send signal (should not occur); do NOT
+         * proceed, as something is broken with current identity
+         */
+        log_setid_failure("mod_fcgid PM", "effective uid", 0);
+        _exit(1);
     }
-    rv = apr_proc_kill(&(procnode->proc_id), SIGTERM);
-    if (ap_unixd_config.suexec_enabled && seteuid(ap_unixd_config.user_id) != 0)
-    {
-        kill(getpid(), SIGTERM);
-        return APR_EACCES;
+    rv = apr_proc_kill(&(procnode->proc_id), sig);
+    if (ap_unixd_config.suexec_enabled && seteuid(ap_unixd_config.user_id) != 0) {
+        /* can't drop privileges after signalling (should not occur); do NOT
+         * proceed any further as euid(0)!
+         */
+        log_setid_failure("mod_fcgid PM", "effective uid", ap_unixd_config.user_id);
+        _exit(1);
     }
     return rv;
 }
 
+apr_status_t proc_kill_gracefully(fcgid_procnode *procnode, server_rec *main_server)
+{
+    return proc_kill_internal(procnode, SIGTERM);
+}
+
 apr_status_t proc_kill_force(fcgid_procnode * procnode,
                              server_rec * main_server)
 {
-    apr_status_t rv;
-
-    if (ap_unixd_config.suexec_enabled && seteuid(0) != 0) {
-
-        /* It's fatal error */
-        kill(getpid(), SIGTERM);
-        return APR_EACCES;
-    }
-    rv = apr_proc_kill(&(procnode->proc_id), SIGKILL);
-    if (ap_unixd_config.suexec_enabled && seteuid(ap_unixd_config.user_id) != 0)
-    {
-        kill(getpid(), SIGTERM);
-        return APR_EACCES;
-    }
-    return rv;
+    return proc_kill_internal(procnode, SIGKILL);
 }
 
-apr_status_t
-proc_wait_process(server_rec * main_server, fcgid_procnode * procnode)
+apr_status_t proc_wait_process(server_rec *main_server, fcgid_procnode *procnode)
 {
     apr_status_t rv;
     int exitcode;
@@ -504,8 +509,7 @@ static apr_status_t set_socket_nonblock(int sd)
     return APR_SUCCESS;
 }
 
-apr_status_t
-proc_connect_ipc(fcgid_procnode * procnode, fcgid_ipc * ipc_handle)
+apr_status_t proc_connect_ipc(fcgid_procnode *procnode, fcgid_ipc *ipc_handle)
 {
     fcgid_namedpipe_handle *handle_info;
     struct sockaddr_un unix_addr;
@@ -526,6 +530,9 @@ proc_connect_ipc(fcgid_procnode * procnode, fcgid_ipc * ipc_handle)
     /* Connect to fastcgi server */
     memset(&unix_addr, 0, sizeof(unix_addr));
     unix_addr.sun_family = AF_UNIX;
+
+    /* PM already made this check for truncation */
+    AP_DEBUG_ASSERT(sizeof unix_addr.sun_path > strlen(procnode->socket_path));
     apr_cpystrn(unix_addr.sun_path, procnode->socket_path,
                 sizeof(unix_addr.sun_path));
 
@@ -546,8 +553,7 @@ proc_connect_ipc(fcgid_procnode * procnode, fcgid_ipc * ipc_handle)
                (char *) &on, sizeof(on));
 
     /* Set nonblock option */
-    if ((rv =
-         set_socket_nonblock(handle_info->handle_socket)) != APR_SUCCESS) {
+    if ((rv = set_socket_nonblock(handle_info->handle_socket)) != APR_SUCCESS) {
         ap_log_rerror(APLOG_MARK, APLOG_DEBUG, rv, ipc_handle->request,
                       "mod_fcgid: can't make unix domain socket nonblocking");
         return rv;
@@ -567,8 +573,8 @@ apr_status_t proc_close_ipc(fcgid_ipc * ipc_handle)
     return rv;
 }
 
-apr_status_t proc_read_ipc(fcgid_ipc * ipc_handle, const char *buffer,
-                           apr_size_t * size)
+apr_status_t proc_read_ipc(fcgid_ipc *ipc_handle, const char *buffer,
+                           apr_size_t *size)
 {
     int retcode, unix_socket;
     fcgid_namedpipe_handle *handle_info;
@@ -602,7 +608,8 @@ apr_status_t proc_read_ipc(fcgid_ipc * ipc_handle, const char *buffer,
                       ipc_handle->request,
                       "mod_fcgid: error polling unix domain socket");
         return errno;
-    } else if (retcode == 0) {
+    }
+    else if (retcode == 0) {
         ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0,
                       ipc_handle->request,
                       "mod_fcgid: read data timeout in %d seconds",
@@ -630,7 +637,7 @@ apr_status_t proc_read_ipc(fcgid_ipc * ipc_handle, const char *buffer,
     return errno;
 }
 
-static apr_status_t socket_writev(fcgid_ipc * ipc_handle,
+static apr_status_t socket_writev(fcgid_ipc *ipc_handle,
                                   struct iovec *vec, int nvec,
                                   int *writecnt)
 {
@@ -714,7 +721,7 @@ static apr_status_t socket_writev(fcgid_ipc * ipc_handle,
     return rv;
 }
 
-static apr_status_t writev_it_all(fcgid_ipc * ipc_handle,
+static apr_status_t writev_it_all(fcgid_ipc *ipc_handle,
                                   struct iovec *vec, int nvec)
 {
     apr_size_t bytes_written = 0;
@@ -757,8 +764,8 @@ static apr_status_t writev_it_all(fcgid_ipc * ipc_handle,
 }
 
 #define FCGID_VEC_COUNT 8
-apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
-                            apr_bucket_brigade * output_brigade)
+apr_status_t proc_write_ipc(fcgid_ipc *ipc_handle,
+                            apr_bucket_brigade *output_brigade)
 {
     apr_status_t rv;
     struct iovec vec[FCGID_VEC_COUNT];
@@ -787,7 +794,8 @@ apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
                                FCGID_VEC_COUNT)) != APR_SUCCESS)
                 return rv;
             nvec = 0;
-        } else
+        }
+        else
             nvec++;
     }
 
@@ -800,9 +808,8 @@ apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
     return APR_SUCCESS;
 }
 
-void
-proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
-                     apr_exit_why_e exitwhy, server_rec * main_server)
+void proc_print_exit_info(fcgid_procnode *procnode, int exitcode,
+                          apr_exit_why_e exitwhy, server_rec *main_server)
 {
     const char *diewhy = NULL;
     char signal_info[HUGE_STRING_LEN];
@@ -861,7 +868,8 @@ proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
                              "get unexpected signal %d", signum);
             }
         }
-    } else if (APR_PROC_CHECK_EXIT(exitwhy)) {
+    }
+    else if (APR_PROC_CHECK_EXIT(exitwhy)) {
         apr_snprintf(signal_info, HUGE_STRING_LEN - 1,
                      "terminated by calling exit(), return code: %d",
                      exitcode);
